from liv.models.clip import clip
from liv.models.clip.clip.model import CLIP
import numpy as np
from numpy.core.numeric import full
import torch
import torch.nn as nn
from torch.nn.modules.activation import Sigmoid
from torch.nn.modules.linear import Identity
import torchvision
from torchvision import transforms
from pathlib import Path
from torchvision.utils import save_image
import torchvision.transforms as T


class LIV(nn.Module):
    def __init__(self, modelid="RN50", device="cuda",
                       lr=1e-5, weight_decay=0.001,
                       visionweight=1.0, langweight=1.0, clipweight=1.0,
                       gamma=0.98, metric="cos", num_negatives=0,
                       grad_text=True, scratch=False):
        super().__init__()

        self.modelid = modelid
        self.device = device
        self.visionweight = visionweight
        self.langweight = langweight
        self.clipweight = clipweight

        self.gamma = gamma
        self.num_negatives = num_negatives
        self.metric = metric
        self.grad_text = grad_text

        # Load CLIP model and transform
        model, cliptransforms = clip.load(modelid, device=self.device, scratch=scratch, jit=False)

        # CLIP precision
        if device == "cpu":
            model.float()
        else :
            clip.model.convert_weights(model)

        self.model = model
        self.model.train()
        self.transforms = cliptransforms

        self.transforms_tensor = nn.Sequential(
                transforms.Resize(self.model.visual.input_resolution, antialias=None),
                transforms.CenterCrop(self.model.visual.input_resolution),
                transforms.Normalize((0.48145466, 0.4578275, 0.40821073), (0.26862954, 0.26130258, 0.27577711))
            )

        self.output_dim = self.model.visual.output_dim

        ## Optimizer
        self.encoder_opt = torch.optim.Adam(list(self.model.parameters()),
        lr=lr, betas=(0.9,0.98),eps=1e-6, weight_decay=weight_decay)

    ## Forward Call (im --> representation)
    def forward(self, input, modality="vision", normalize=True):
        if modality == "vision":
            if type(input) != torch.Tensor:
                print("Warning: Input not tensor, may cause issue with normalization")
                input = self.transforms(input).to(self.device)
            else:
                # rescale to [0, 1]
                if torch.max(input) > 10.0:
                    input = input / 255.0
                input = self.transforms_tensor(input).to(self.device)

            features = self.model.encode_image(input)
        elif modality == "text":
            b_token = input
            if self.grad_text:
                features = self.model.encode_text(b_token)
            else:
                with torch.no_grad():
                    features = self.model.encode_text(b_token)
        else:
            raise NotImplementedError

        return features

    def encode_image_as_features(self, img):
        img = self.transforms_tensor(img).to(self.device)
        # [B,C,H,W]
        return self.model.encode_image_as_features(img)

    def encode_text_as_embeddings(self, text):
        return self.model.encode_text_as_embeddings(text)

    def sim(self, tensor1, tensor2):
        if type(tensor1) == np.ndarray:
            tensor1 = torch.from_numpy(tensor1).to(self.device)
            tensor2 = torch.from_numpy(tensor2).to(self.device)
        if self.metric == 'l2':
            d = -torch.linalg.norm(tensor1 - tensor2, dim = -1)
        elif self.metric == 'cos':
            tensor1 = tensor1 / tensor1.norm(dim=-1, keepdim=True)
            tensor2 = tensor2 / tensor2.norm(dim=-1, keepdim=True)
            d = torch.nn.CosineSimilarity(-1)(tensor1, tensor2)
        else:
            raise NotImplementedError
        return d

    def get_reward(self, e0, es, le, encoded=True):
        assert encoded == True
        return self.sim(es, le)




import sys
import os
import glob
import cv2
import torch
import numpy as np
from PIL import Image
import torchvision.transforms as T
import clip
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation, PillowWriter

from liv import load_liv


def plot_rewards(distances_cur_img, distances_cur_text, imgs, task, fig_filename, animated=False):
    fig, ax = plt.subplots(nrows=1, ncols=4, figsize=(24, 6))
    ax[0].plot(
        np.arange(len(distances_cur_img)),
        distances_cur_img,
        color="tab:blue",
        label="image",
        linewidth=3,
    )
    ax[0].plot(
        np.arange(len(distances_cur_text)),
        distances_cur_text,
        color="tab:red",
        label="text",
        linewidth=3,
    )
    ax[1].plot(
        np.arange(len(distances_cur_img)),
        distances_cur_img,
        color="tab:blue",
        label="image",
        linewidth=3,
    )
    ax[2].plot(
        np.arange(len(distances_cur_text)),
        distances_cur_text,
        color="tab:red",
        label="text",
        linewidth=3,
    )
    ax[0].legend(loc="upper right")
    ax[0].set_xlabel("Frame", fontsize=15)
    ax[1].set_xlabel("Frame", fontsize=15)
    ax[2].set_xlabel("Frame", fontsize=15)
    ax[0].set_ylabel("Embedding Distance", fontsize=15)
    ax[0].set_title(f"Language Goal: {task}", fontsize=15)
    ax[1].set_title("Image Goal", fontsize=15)
    ax[2].set_title(f"Language Goal: {task}", fontsize=15)
    ax[3].imshow(imgs[-1].permute(1, 2, 0))
    asp = 1
    ax[0].set_aspect(asp * np.diff(ax[0].get_xlim())[0] / np.diff(ax[0].get_ylim())[0])
    ax[1].set_aspect(asp * np.diff(ax[1].get_xlim())[0] / np.diff(ax[1].get_ylim())[0])
    ax[2].set_aspect(asp * np.diff(ax[2].get_xlim())[0] / np.diff(ax[2].get_ylim())[0])
    fig.savefig(f"{fig_filename}.png", bbox_inches="tight")
    plt.close()

    ax0_xlim = ax[0].get_xlim()
    ax0_ylim = ax[0].get_ylim()
    ax1_xlim = ax[1].get_xlim()
    ax1_ylim = ax[1].get_ylim()
    ax2_xlim = ax[2].get_xlim()
    ax2_ylim = ax[2].get_ylim()

    def animate(i):
        i = min(i, len(distances_cur_img) - 1)
        for ax_subplot in ax:
            ax_subplot.clear()
        ranges = np.arange(len(distances_cur_img))
        line1 = ax[0].plot(
            ranges[0:i], distances_cur_img[0:i], color="tab:blue", label="image", linewidth=3
        )
        line2 = ax[0].plot(
            ranges[0:i], distances_cur_text[0:i], color="tab:red", label="text", linewidth=3
        )
        line3 = ax[1].plot(
            ranges[0:i], distances_cur_img[0:i], color="tab:blue", label="image", linewidth=3
        )
        line4 = ax[2].plot(
            ranges[0:i], distances_cur_text[0:i], color="tab:red", label="text", linewidth=3
        )
        line5 = ax[3].imshow(imgs[i].permute(1, 2, 0))
        ax[0].legend(loc="upper right")
        ax[0].set_xlabel("Frame", fontsize=15)
        ax[1].set_xlabel("Frame", fontsize=15)
        ax[2].set_xlabel("Frame", fontsize=15)
        ax[0].set_ylabel("Embedding Distance", fontsize=15)
        ax[0].set_title(f"Language Goal: {task}", fontsize=15)
        ax[1].set_title("Image Goal", fontsize=15)
        ax[2].set_title(f"Language Goal: {task}", fontsize=15)

        ax[0].set_xlim(ax0_xlim)
        ax[0].set_ylim(ax0_ylim)
        ax[1].set_xlim(ax1_xlim)
        ax[1].set_ylim(ax1_ylim)
        ax[2].set_xlim(ax2_xlim)
        ax[2].set_ylim(ax2_ylim)

        return line1, line2, line3, line4, line5

    if animated:
        final_freeze = 30
        ani = FuncAnimation(
            fig, animate, interval=20, repeat=False, frames=(len(distances_cur_img) + final_freeze)
        )
        ani.save(f"{fig_filename}.gif", dpi=100, writer=PillowWriter(fps=25))


def calculate_distances(encoder_model, imgs, task):
    with torch.no_grad():
        embeddings = encoder_model(input=imgs.cuda(), modality="vision")
        goal_embedding_img = embeddings[-1]
        token = clip.tokenize([task])
        goal_embedding_text = encoder_model(input=token, modality="text")
        goal_embedding_text = goal_embedding_text[0]

    distances_cur_img = []
    distances_cur_text = []
    for t in range(embeddings.shape[0]):
        cur_embedding = embeddings[t]
        cur_distance_img = (
            -encoder_model.module.sim(goal_embedding_img, cur_embedding).detach().cpu().numpy()
        )
        cur_distance_text = (
            -encoder_model.module.sim(goal_embedding_text, cur_embedding).detach().cpu().numpy()
        )
        distances_cur_img.append(cur_distance_img)
        distances_cur_text.append(cur_distance_text)

    distances_cur_img = np.array(distances_cur_img)
    distances_cur_text = np.array(distances_cur_text)

    return distances_cur_img, distances_cur_text


def plot_reward_curves(
    manifest, tasks, load_video, encoder_model, fig_filename_prefix, animated=False, num_vid=5
):
    for task in tasks:
        try:
            videos = manifest[manifest["text"] == task]
        except:
            videos  = manifest[manifest["narration"] == task]
        for i in range(num_vid):
            m = videos.iloc[i]
            imgs = load_video(m) 
            fig_filename = f"{fig_filename_prefix}_{task}_{i}".replace(" ", "-")
            distances_cur_img, distances_cur_text = calculate_distances(
                encoder_model, imgs, task
            )
            plot_rewards(
                distances_cur_img,
                distances_cur_text,
                imgs,
                task,
                fig_filename,
                animated=animated,
            )



FRANKAKITCHEN_MAPPING = {
    'kitchen_micro_open-v3' : "open microwave",
    'kitchen_sdoor_open-v3' : "slide cabinet",
    'kitchen_ldoor_open-v3' : "open left door",
    'kitchen_light_on-v3' : "switch on light", 
    'kitchen_knob1_on-v3': "turn on stove",
}


import clip
import numpy as np
from numpy.core.numeric import full
import torch
import torch.nn as nn
import torch.nn.functional as F
import random
from pathlib import Path
from torchvision.utils import save_image
import time
import copy
import torchvision.transforms as T

epsilon = 1e-8
def do_nothing(x): return x

class Trainer():
    def __init__(self):
        self.clip_loss_img = nn.CrossEntropyLoss()
        self.clip_loss_txt = nn.CrossEntropyLoss()
    
    # A simplified function call of CLIP.forward()
    def compute_clip_loss(self, model, image_features, text_features):
        image_features = image_features / image_features.norm(dim=1, keepdim=True)
        text_features = text_features / text_features.norm(dim=1, keepdim=True)

        logits_per_image = image_features @ text_features.t()
        logits_per_text = logits_per_image.t()
        ground_truth = torch.arange(len(image_features),dtype=torch.long,device=image_features.device)
        clip_loss = (self.clip_loss_img(logits_per_image,ground_truth) + self.clip_loss_txt(logits_per_text,ground_truth))/2
        return clip_loss 

    def compute_vip_loss(self, model, e0, es0_vip, es1_vip, eg, b_reward, num_negatives=0):
        r =  b_reward.to(e0.device)

        V_0 = model.module.sim(e0, eg)
        V_s = model.module.sim(es0_vip, eg)
        V_s_next = model.module.sim(es1_vip, eg)

        # Rescale Value 
        V_0 = V_0 / (1-model.module.gamma)
        V_s = V_s / (1-model.module.gamma)
        V_s_next = V_s_next / (1-model.module.gamma)

        # Compute VIP Loss
        V_loss = (1-model.module.gamma) * -V_0.mean() + torch.log(epsilon + torch.mean(torch.exp(-(r + model.module.gamma * V_s_next - V_s))))

        # Optionally, add additional "negative" observations
        if num_negatives > 0:
            V_s_neg = []
            V_s_next_neg = []
            for _ in range(num_negatives):
                perm = torch.randperm(es0_vip.size()[0])
                es0_vip_shuf = es0_vip[perm]
                es1_vip_shuf = es1_vip[perm]

                V_s_neg.append(model.module.sim(es0_vip_shuf, eg))
                V_s_next_neg.append(model.module.sim(es1_vip_shuf, eg))

            V_s_neg = torch.cat(V_s_neg)
            V_s_next_neg = torch.cat(V_s_next_neg)
            r_neg = -torch.ones(V_s_neg.shape).to(V_0.device)
            V_s_neg = V_s_neg / (1-model.module.gamma)
            V_s_next_neg = V_s_next_neg / (1-model.module.gamma)                
            V_loss = V_loss + torch.log(epsilon + torch.mean(torch.exp(-(r_neg + model.module.gamma * V_s_next_neg - V_s_neg))))
        
        return V_loss 

    def update(self, model, batch, step, eval=False):
        t0 = time.time()
        metrics = dict()
        if eval:
            model.eval()
        else:
            model.train()

        ## Batch
        b_im, b_reward, b_lang = batch
        b_im = b_im.cuda()
        bs = b_im.shape[0]
        img_stack_size = b_im.shape[1]
        H = b_im.shape[-2]
        W = b_im.shape[-1]
        b_im_r = b_im.reshape(bs*img_stack_size, 3, H, W)

        # Encode visual and text inputs
        e_img = model(b_im_r, modality="vision")
        b_token = clip.tokenize(b_lang)
        e_lang = model(b_token, modality="text")

        e_img = e_img.reshape(bs, img_stack_size, -1)
        e0 = e_img[:, 0] # initial, o_0
        eg = e_img[:, 1] # final, o_g
        es0_vip = e_img[:, 2] # o_t
        es1_vip = e_img[:, 3] # o_t+1
        eg_img = e_img[:, -1] 
        full_loss = 0

        ## CLIP Loss 
        if model.module.clipweight != 0:
            clip_loss = self.compute_clip_loss(model, eg_img, e_lang)
            clip_loss = model.module.clipweight * clip_loss
            metrics['clip_loss'] = clip_loss.item()
            full_loss += clip_loss

        ## VIP Loss (Visual)
        vip_loss_visual = self.compute_vip_loss(model, e0, es0_vip, es1_vip, eg, b_reward, model.module.num_negatives)
        metrics['vip_loss_visual'] = vip_loss_visual.item()
        full_loss += model.module.visionweight * vip_loss_visual

        ## VIP Loss (Language)
        if model.module.langweight != 0:
            vip_loss_lang = self.compute_vip_loss(model, e0, es0_vip, es1_vip, e_lang, b_reward, model.module.num_negatives)
            metrics['vip_loss_lang'] = vip_loss_lang.item()
            full_loss += model.module.langweight * vip_loss_lang

        metrics['full_loss'] = full_loss.item()

        if not eval:
            model.module.encoder_opt.zero_grad()
            full_loss.backward()
            model.module.encoder_opt.step()

        return metrics, None



import csv
import datetime
from collections import defaultdict

import numpy as np
import torch
import torchvision
import wandb

COMMON_TRAIN_FORMAT = [('frame', 'F', 'int'), ('step', 'S', 'int'),
                       ('episode', 'E', 'int'), ('episode_length', 'L', 'int'),
                       ('episode_reward', 'R', 'float'),
                       ('buffer_size', 'BS', 'int'), ('fps', 'FPS', 'float'),
                       ('total_time', 'T', 'time')]

COMMON_EVAL_FORMAT = [('frame', 'F', 'int'), ('step', 'S', 'int'),
                      ('episode', 'E', 'int'), ('episode_length', 'L', 'int'),
                      ('episode_reward', 'R', 'float'),
                      ('total_time', 'T', 'time')]


class AverageMeter(object):
    def __init__(self):
        self._sum = 0
        self._count = 0

    def update(self, value, n=1):
        self._sum += value
        self._count += n

    def value(self):
        return self._sum / max(1, self._count)


class MetersGroup(object):
    def __init__(self, csv_file_name, formating):
        self._csv_file_name = csv_file_name
        self._formating = formating
        self._meters = defaultdict(AverageMeter)
        self._csv_file = None
        self._csv_writer = None

    def log(self, key, value, n=1):
        self._meters[key].update(value, n)

    def _prime_meters(self):
        data = dict()
        for key, meter in self._meters.items():
            if key.startswith('train'):
                key = key[len('train') + 1:]
            else:
                key = key[len('eval') + 1:]
            key = key.replace('/', '_')
            data[key] = meter.value()
        return data

    def _remove_old_entries(self, data):
        rows = []
        with self._csv_file_name.open('r') as f:
            reader = csv.DictReader(f)
            for row in reader:
                if float(row['episode']) >= data['episode']:
                    break
                rows.append(row)
        with self._csv_file_name.open('w') as f:
            writer = csv.DictWriter(f,
                                    fieldnames=sorted(data.keys()),
                                    restval=0.0)
            writer.writeheader()
            for row in rows:
                writer.writerow(row)

    def _dump_to_csv(self, data):
        if self._csv_writer is None:
            should_write_header = True
            if self._csv_file_name.exists():
                self._remove_old_entries(data)
                should_write_header = False

            self._csv_file = self._csv_file_name.open('a')
            self._csv_writer = csv.DictWriter(self._csv_file,
                                              fieldnames=sorted(data.keys()),
                                              restval=0.0)
            if should_write_header:
                self._csv_writer.writeheader()

        self._csv_writer.writerow(data)
        self._csv_file.flush()

    def _format(self, key, value, ty):
        if ty == 'int':
            value = int(value)
            return f'{key}: {value}'
        elif ty == 'float':
            return f'{key}: {value:.04f}'
        elif ty == 'time':
            value = str(datetime.timedelta(seconds=int(value)))
            return f'{key}: {value}'
        else:
            raise f'invalid format type: {ty}'

    def _dump_to_console(self, data, prefix):
        prefix = colored(prefix, 'yellow' if prefix == 'train' else 'green')
        pieces = [f'| {prefix: <14}']
        for key, disp_key, ty in self._formating:
            value = data.get(key, 0)
            pieces.append(self._format(disp_key, value, ty))
        print(' | '.join(pieces))

    def dump(self, step, prefix):
        if len(self._meters) == 0:
            return
        data = self._prime_meters()
        data['frame'] = step
        self._dump_to_csv(data)
        self._dump_to_console(data, prefix)
        self._meters.clear()


class Logger(object):
    def __init__(self, log_dir, use_tb, cfg=None):
        self._log_dir = log_dir
        self.use_tb = use_tb
        self._train_mg = MetersGroup(log_dir / 'train.csv',
                                     formating=COMMON_TRAIN_FORMAT)
        self._eval_mg = MetersGroup(log_dir / 'eval.csv',
                                    formating=COMMON_EVAL_FORMAT)
        if use_tb:
            self._sw = SummaryWriter(str(log_dir / 'tb'))
        else:
            wandb.init(project=cfg.wandbproject, entity=cfg.wandbuser, name=cfg.experiment)
            fullcfg = {**cfg, **cfg.agent}
            wandb.config.update(fullcfg)

    def _try_sw_log(self, key, value, step):
        if self.use_tb:
            self._sw.add_scalar(key, value, step)
        else:
            logs = {}
            logs[key] = value
            wandb.log(logs, step = step)

    def log(self, key, value, step):
        assert key.startswith('train') or key.startswith('eval')
        if type(value) == torch.Tensor:
            value = value.item()
        self._try_sw_log(key, value, step)
        mg = self._train_mg if key.startswith('train') else self._eval_mg
        mg.log(key, value)

    def log_metrics(self, metrics, step, ty):
        for key, value in metrics.items():
            self.log(f'{ty}/{key}', value, step)

    def dump(self, step, ty=None):
        if ty is None or ty == 'eval':
            self._eval_mg.dump(step, 'eval')
        if ty is None or ty == 'train':
            self._train_mg.dump(step, 'train')

    def log_and_dump_ctx(self, step, ty):
        return LogAndDumpCtx(self, step, ty)


class LogAndDumpCtx:
    def __init__(self, logger, step, ty):
        self._logger = logger
        self._step = step
        self._ty = ty

    def __enter__(self):
        return self

    def __call__(self, key, value):
        self._logger.log(f'{self._ty}/{key}', value, self._step)

    def __exit__(self, *args):
        self._logger.dump(self._step, self._ty)



import os
from os.path import expanduser
import omegaconf
import hydra
from huggingface_hub import hf_hub_download
import gdown
import torch
import copy
from liv.models.model_liv import LIV

VALID_ARGS = ["_target_", "device", "lr", "hidden_dim", "size", "l2weight", "l1weight", "num_negatives"]

if torch.cuda.is_available():
    device = "cuda"
else:
    device = "cpu"

def cleanup_config(cfg):
    config = copy.deepcopy(cfg)
    config["device"] = device
    return config.agent

def load_liv(modelid='resnet50'):
    assert modelid == 'resnet50'
    home = os.path.join(expanduser("~"), ".liv")

    if not os.path.exists(os.path.join(home, modelid)):
        os.makedirs(os.path.join(home, modelid))
    folderpath = os.path.join(home, modelid)
    modelpath = os.path.join(home, modelid, "model.pt")
    configpath = os.path.join(home, modelid, "config.yaml")

    if not os.path.exists(modelpath):
        try:
            # Default reliable download from HuggingFace Hub
            hf_hub_download(repo_id="jasonyma/LIV", filename="model.pt", local_dir=folderpath)
            hf_hub_download(repo_id="jasonyma/LIV", filename="config.yaml", local_dir=folderpath)
        except:
            # Download from GDown
            modelurl = 'https://drive.google.com/uc?id=1l1ufzVLxpE5BK7JY6ZnVBljVzmK5c4P3'
            configurl = 'https://drive.google.com/uc?id=1GWA5oSJDuHGB2WEdyZZmkro83FNmtaWl'
            gdown.download(modelurl, modelpath, quiet=False)
            gdown.download(configurl, configpath, quiet=False)

    modelcfg = omegaconf.OmegaConf.load(configpath)
    cleancfg = cleanup_config(modelcfg)
    rep = hydra.utils.instantiate(cleancfg)
    #rep = torch.nn.DataParallel(rep)
    state_dict = torch.load(modelpath, map_location=torch.device(device))['liv']
    # Remove the prefix '.module' as the paras are stored as a DP model
    new_state_dict = {k.replace("module.", ''): v for k, v in state_dict.items()}
    rep.load_state_dict(new_state_dict)
    return rep




import warnings
warnings.filterwarnings('ignore', category=DeprecationWarning)

import os
os.environ['MKL_SERVICE_FORCE_INTEL'] = '1'
os.environ['MUJOCO_GL'] = 'egl'

from pathlib import Path
from PIL import Image  
import clip 
import glob 
import hydra
from matplotlib import pyplot as plt 
from matplotlib.animation import FuncAnimation, PillowWriter
import numpy as np
import pandas as pd 
import torch
import torchvision.transforms as T
import time

from liv import load_liv
from liv.trainer import Trainer
from liv.utils import utils
from liv.utils.data_loaders import LIVBuffer
from liv.utils.logger import Logger
from liv.utils.plotter import plot_reward_curves

def make_network(cfg):
    model =  hydra.utils.instantiate(cfg)
    print("Let's use", torch.cuda.device_count(), "GPUs!")
    model = torch.nn.DataParallel(model)
    if cfg.device == "cpu":
        model = model.module.to(cfg.device)
    return model

class Workspace:
    def __init__(self, cfg):
        self.work_dir = Path.cwd()
        print(f'workspace: {self.work_dir}')

        self.cfg = cfg
        self.logging = self.cfg.logging
        utils.set_seed_everywhere(cfg.seed)
        self.device = torch.device(cfg.device)
        if self.logging:
            self.setup()

        if not cfg.eval:
            print("Creating Dataloader")
            train_iterable = LIVBuffer(datasource=self.cfg.dataset, datapath=self.cfg.datapath_train, num_workers=self.cfg.num_workers, num_demos=self.cfg.num_demos, doaug=self.cfg.doaug, alpha=self.cfg.alpha)
            self.train_dataset = train_iterable
            self.train_loader = iter(torch.utils.data.DataLoader(train_iterable,
                                            batch_size=self.cfg.batch_size,
                                            num_workers=self.cfg.num_workers,
                                            pin_memory=True))

        ## Init Model
        print("Initializing Model")
        self.model = make_network(cfg.agent)
        self.timer = utils.Timer()
        self._global_step = 0

        ## If reloading existing model
        if cfg.load_snap:
            print("LOADING", cfg.load_snap)
            self.load_snapshot(cfg.load_snap)

    def setup(self):
        # create logger
        self.logger = Logger(self.work_dir, use_tb=False, cfg=self.cfg)

    @property
    def global_step(self):
        return self._global_step

    @property
    def global_frame(self):
        return self.global_step

    def train(self):
        # predicates
        train_until_step = utils.Until(self.cfg.train_steps, 1)
        eval_freq = self.cfg.eval_freq
        eval_every_step = utils.Every(eval_freq, 1)

        # trainer = Trainer()
        trainer = hydra.utils.instantiate(self.cfg.trainer)

        ## Training Loop
        print("Begin Training")
        while train_until_step(self.global_step):
            if eval_every_step(self.global_step):
                self.generate_reward_curves()
                self.save_snapshot()
            
            ## Sample Batch
            t0 = time.time()
            batch = next(self.train_loader)
            t1 = time.time()
            metrics, st = trainer.update(self.model, batch, self.global_step)
            t2 = time.time()
            if self.logging:
                self.logger.log_metrics(metrics, self.global_frame, ty='train')

            if self.global_step % 10 == 0:
                print(self.global_step, metrics)
                print(f'Sample time {t1-t0}, Update time {t2-t1}')
                
            self._global_step += 1

    def save_snapshot(self):
        snapshot = self.work_dir / f'snapshot_{self.global_step}.pt'
        global_snapshot =  self.work_dir / f'snapshot.pt'
        sdict = {}
        sdict["liv"] = self.model.module.state_dict()
        sdict["optimizer"] = self.model.module.encoder_opt.state_dict()
        sdict["global_step"] = self._global_step
        torch.save(sdict, snapshot)
        torch.save(sdict, global_snapshot)

    def load_snapshot(self, snapshot_path):
        if snapshot_path != 'liv':
            payload = torch.load(snapshot_path)
            self.model.module.load_state_dict(payload['liv'])
        else:
            self.model = load_liv()
        clip.model.convert_weights(self.model)
        try:
            self._global_step = payload['global_step']
        except:
            print("Warning: No global step found")

    def generate_reward_curves(self):
        self.model.eval()
        os.makedirs(f"{self.work_dir}/reward_curves", exist_ok=True)
        transform = T.Compose([T.ToTensor()])

        if self.cfg.dataset not in ["epickitchen"]:
            manifest = pd.read_csv(os.path.join(self.cfg.datapath_train, "manifest.csv"))
            tasks = manifest["text"].unique()
        else:
            manifest = pd.read_csv(os.path.join(self.cfg.datapath_train, "EPIC_100_validation.csv"))
            tasks = ["open microwave", "open cabinet", "open door"]

        fig_filename = f"{self.work_dir}/reward_curves/{self._global_step}_{self.cfg.dataset}"
        if self.cfg.dataset in ["epickitchen"]:
            def load_video(m):
                imgs_tensor = []
                start_frame = m["start_frame"]
                end_frame = m["stop_frame"]
                vid = f"/data2/jasonyma/EPIC-KITCHENS/frames/{m['participant_id']}/rgb_frames/{m['video_id']}"
                for index in range(start_frame, end_frame):
                    img = Image.open(f"{vid}/frame_0000{index+1:06}.jpg")
                    imgs_tensor.append(transform(img))
                imgs_tensor = torch.stack(imgs_tensor)
                return imgs_tensor

        else:
            def load_video(m):
                imgs_tensor = []
                vid = m["directory"]
                for index in range(m["num_frames"]):
                    try:
                        img = Image.open(f"{vid}/{index}.png")
                    except:
                        img = Image.open(f"{vid}/{index}.jpg")
                    imgs_tensor.append(transform(img))
                imgs_tensor = torch.stack(imgs_tensor)
                return imgs_tensor

        plot_reward_curves(
            manifest,
            tasks,
            load_video,
            self.model,
            fig_filename,
            animated=self.cfg.animate,
        )
        self.model.train()


@hydra.main(config_path='cfgs', config_name='config_liv')
def main(cfg):
    from liv.train_liv import Workspace as W
    root_dir = Path.cwd()
    workspace = W(cfg)

    snapshot = root_dir / 'snapshot.pt'
    if snapshot.exists():
        print(f'resuming: {snapshot}')
        workspace.load_snapshot(snapshot)

    if not cfg.eval:
        workspace.train()
    else:
        workspace.generate_reward_curves()

if __name__ == '__main__':
    main()


import random
import re
import time

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from omegaconf import OmegaConf
from torch import distributions as pyd
from torch.distributions.utils import _standard_normal


class eval_mode:
    def __init__(self, *models):
        self.models = models

    def __enter__(self):
        self.prev_states = []
        for model in self.models:
            self.prev_states.append(model.training)
            model.train(False)

    def __exit__(self, *args):
        for model, state in zip(self.models, self.prev_states):
            model.train(state)
        return False


def set_seed_everywhere(seed):
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)


def soft_update_params(net, target_net, tau):
    for param, target_param in zip(net.parameters(), target_net.parameters()):
        target_param.data.copy_(tau * param.data +
                                (1 - tau) * target_param.data)


def to_torch(xs, device):
    return tuple(torch.as_tensor(x, device=device) for x in xs)


def weight_init(m):
    if isinstance(m, nn.Linear):
        nn.init.orthogonal_(m.weight.data)
        if hasattr(m.bias, 'data'):
            m.bias.data.fill_(0.0)
    elif isinstance(m, nn.Conv2d) or isinstance(m, nn.ConvTranspose2d):
        gain = nn.init.calculate_gain('relu')
        nn.init.orthogonal_(m.weight.data, gain)
        if hasattr(m.bias, 'data'):
            m.bias.data.fill_(0.0)

def accuracy(output, target, topk=(1,)):
    """Computes the precision@k for the specified values of k"""
    maxk = max(topk)
    batch_size = target.size(0)

    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in topk:
        correct_k = correct[:k].reshape(-1).float().sum(0)
        res.append(correct_k.mul_(1.0 / batch_size))
    return res

class Until:
    def __init__(self, until, action_repeat=1):
        self._until = until
        self._action_repeat = action_repeat

    def __call__(self, step):
        if self._until is None:
            return True
        until = self._until // self._action_repeat
        return step < until


class Every:
    def __init__(self, every, action_repeat=1):
        self._every = every
        self._action_repeat = action_repeat

    def __call__(self, step):
        if self._every is None:
            return False
        every = self._every // self._action_repeat
        if step % every == 0:
            return True
        return False


class Timer:
    def __init__(self):
        self._start_time = time.time()
        self._last_time = time.time()

    def reset(self):
        elapsed_time = time.time() - self._last_time
        self._last_time = time.time()
        total_time = time.time() - self._start_time
        return elapsed_time, total_time

    def total_time(self):
        return time.time() - self._start_time


class TruncatedNormal(pyd.Normal):
    def __init__(self, loc, scale, low=-1.0, high=1.0, eps=1e-6):
        super().__init__(loc, scale, validate_args=False)
        self.low = low
        self.high = high
        self.eps = eps

    def _clamp(self, x):
        clamped_x = torch.clamp(x, self.low + self.eps, self.high - self.eps)
        x = x - x.detach() + clamped_x.detach()
        return x

    def sample(self, clip=None, sample_shape=torch.Size()):
        shape = self._extended_shape(sample_shape)
        eps = _standard_normal(shape,
                               dtype=self.loc.dtype,
                               device=self.loc.device)
        eps *= self.scale
        if clip is not None:
            eps = torch.clamp(eps, -clip, clip)
        x = self.loc + eps
        return self._clamp(x)


def schedule(schdl, step):
    try:
        return float(schdl)
    except ValueError:
        match = re.match(r'linear\((.+),(.+),(.+)\)', schdl)
        if match:
            init, final, duration = [float(g) for g in match.groups()]
            mix = np.clip(step / duration, 0.0, 1.0)
            return (1.0 - mix) * init + mix * final
        match = re.match(r'step_linear\((.+),(.+),(.+),(.+),(.+)\)', schdl)
        if match:
            init, final1, duration1, final2, duration2 = [
                float(g) for g in match.groups()
            ]
            if step <= duration1:
                mix = np.clip(step / duration1, 0.0, 1.0)
                return (1.0 - mix) * init + mix * final1
            else:
                mix = np.clip((step - duration1) / duration2, 0.0, 1.0)
                return (1.0 - mix) * final1 + mix * final2
    raise NotImplementedError(schdl)



import warnings

import torchvision
warnings.filterwarnings('ignore', category=DeprecationWarning)

import os
from pathlib import Path
import glob
import numpy as np
import torch
from torchvision import transforms
from torch.utils.data import IterableDataset
import pandas as pd
import json
import time
import pickle
from torchvision.utils import save_image
import json
import re
import random

def tryint(s):
    try:
        return int(s)
    except:
        return s

def alphanum_key(s):
    """ Turn a string into a list of string and number chunks.
        "z23a" -> ["z", 23, "a"]
    """
    return [ tryint(c) for c in re.split('([0-9]+)', s) ]

def sort_nicely(l):
    """ Sort the given list in the way that humans expect.
    """
    l.sort(key=alphanum_key)
    return l 

def get_ind(vid, index, ds="ego4d"):
    if "epickitchen" in ds:
        return torchvision.io.read_image(f"{vid}/frame_0000{index+1:06}.jpg")
    else:
        try:
            return torchvision.io.read_image(os.path.join(vid, f"{index}.jpg"))
        except: 
            return torchvision.io.read_image(os.path.join(vid, f"{index}.png"))

## Data Loader for LIV
class LIVBuffer(IterableDataset):
    def __init__(self, datasource='epickitchen', datapath=None, num_workers=8, num_demos=100, doaug = "none", alpha=0.95):
        self._num_workers = max(1, num_workers)
        self.num_demos = num_demos
        self.alpha = alpha
        self.datasource = datasource
        self.datapath = datapath
        assert(datapath is not None)
        self.doaug = doaug
        
        # augmentations
        self.preprocess = torch.nn.Sequential(
                        transforms.Resize(256, antialias=None),
                        transforms.CenterCrop(224)
                )

        # default data augmentation 
        if doaug in ["rc", "rctraj"]:
            self.aug = torch.nn.Sequential(
                transforms.RandomResizedCrop(224, scale = (0.2, 1.0), antialias=None),
            )
        elif doaug in ["metaworld"]:
            self.aug = torch.nn.Sequential(
                transforms.ColorJitter(brightness=0.02, contrast=0.02, saturation=0.02, hue=0.02),
                transforms.RandomAffine(20, translate=(0.1, 0.1), scale=(0.9, 1.1)),
            )
        else:
            self.aug = lambda a : a

        # Load data
        if "epickitchen" in self.datasource: 
            self.manifest = pd.read_csv(f"{self.datapath}/EPIC_100_train.csv")
            self.datalen = len(self.manifest)
        else:
            self.manifest = pd.read_csv(f"{self.datapath}/manifest.csv")
            self.datalen = len(self.manifest)

    def _sample(self):
        # Sample a video from datasource
        if self.datasource == 'epickitchen':
            vidid = np.random.randint(0, self.datalen)
            m = self.manifest.iloc[vidid]
            start_frame = m["start_frame"]
            end_frame = m["stop_frame"]
            vidlen = end_frame - start_frame + 1
            text = m["narration"]
            vid = f"/data2/jasonyma/EPIC-KITCHENS/frames/{m['participant_id']}/rgb_frames/{m['video_id']}"
        else: 
            vidid = np.random.randint(0, self.datalen)
            m = self.manifest.iloc[vidid]
            vidlen = m["num_frames"]
            text = m["text"]
            vid = m["directory"]
            start_frame = 0
            end_frame = vidlen

        # Sample (o_t, o_k, o_k+1, o_T) for LIV training
        if self.datasource in ["epickitchen"]:
            start_ind = np.random.randint(0, vidlen-2)
            end_ind = np.random.randint(start_ind+1, vidlen)
            end_text_ind = np.random.randint(int(self.alpha * vidlen) - 1, vidlen)
        else:
            start_ind = 0
            end_ind = vidlen - 1
            end_text_ind = vidlen - 1

        start_ind = start_ind + start_frame 
        end_ind = end_ind + start_frame 
        end_text_ind = end_text_ind + start_frame 

        s0_ind = np.random.randint(start_ind, end_ind)
        s1_ind = min(s0_ind+1, end_ind)

        # Self-supervised reward (this is always -1)
        reward = float(s0_ind == end_ind) - 1 

        if self.doaug == "rctraj":
            ### Encode each image in the video at once the same way
            im0 = get_ind(vid, start_ind, self.datasource) 
            img = get_ind(vid, end_ind, self.datasource)
            img_text = get_ind(vid, end_text_ind, self.datasource)
            imts0 = get_ind(vid, s0_ind, self.datasource)
            imts1 = get_ind(vid, s1_ind, self.datasource)

            allims = torch.stack([im0, img, imts0, imts1, img_text], 0)
            allims_aug = self.aug(allims / 255.0) 

            im0 = allims_aug[0]
            img = allims_aug[1]
            imts0 = allims_aug[2]
            imts1 = allims_aug[3]
            img_text = allims_aug[4]
        else:
            ### Encode each image individually
            im0 = self.aug(get_ind(vid, start_ind, self.datasource) / 255.0) 
            img = self.aug(get_ind(vid, end_ind, self.datasource) / 255.0)
            imts0 = self.aug(get_ind(vid, s0_ind, self.datasource) / 255.0) 
            imts1 = self.aug(get_ind(vid, s1_ind, self.datasource) / 255.0) 
            img_text = self.aug(get_ind(vid, end_text_ind, self.datasource) / 255.0) 

        im = torch.stack([im0, img, imts0, imts1, img_text])
        return (im, reward, text)

    def __iter__(self):
        while True:
            yield self._sample()

if __name__ == "__main__":
    # You can test your dataset here by replacing the dummy datapath 
    datasource = ""
    datapath = ""
    buffer = LIVBuffer(datasource=datasource, datapath=datapath, doaug='rctraj')
    train_loader = iter(torch.utils.data.DataLoader(buffer, batch_size=2, num_workers=1, pin_memory=True))
    for i in range(1):
        img, reward, text, action = next(train_loader)
        print(img.shape, reward.shape, action.shape, text)


import os
import sys
from setuptools import setup, find_packages

if sys.version_info.major != 3:
    print("This Python is only compatible with Python 3, but you are running "
          "Python {}. The installation will likely fail.".format(sys.version_info.major))

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='liv',
    version='0.1.3',
    packages=find_packages(),
    description='LIV: Language-Image Representations and Rewards for Robotic Control',
    long_description=read('README.md'),
    author='Jason Ma',
    install_requires=[
        'torch',
        'torchvision>=0.8.2',
        'omegaconf==2.1.1',
        'hydra-core==1.1.1',
        'pillow==9.5.0', # This version is compatible with tensorboard
        'opencv-python',
        'matplotlib',
        'flatten_dict',
        'gdown',
        'huggingface_hub',
        'tabulate',
        'pandas',
        'scipy',
        'scikit-learn',
        'scikit-video',
        'transforms3d',
        'moviepy',
        'termcolor',
        'ftfy',
        'regex',
        'tqdm'
    ],
    package_dir={"": "."},
    package_data={"": ["*.sh", "*.yaml", "*.txt.gz"]}
)
